#include <stdlib.h>
#include <math.h>

#include <glib.h>
#include <GL/gl.h>

#include <g3dviewer_plugin.h>

double object_radius(G3DObject *object)
{
	int i;
	double max_radius = 0.0;

	g_return_val_if_fail(object != NULL, 0.0);

	for(i=0; i<object->vertex_count; i++)
	{
		GLfloat *v = &object->vertices[i*3];
		double r = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
		if(r > max_radius) max_radius = r;
	}
#if DEBUG > 2
	g_printerr("cnt: %d, max_radius: %f, sqrt(): %f\n", object->vertex_count,
		max_radius, sqrt(max_radius));
#endif
	return sqrt(max_radius);
}

void object_scale(G3DObject *object, GLfloat scale)
{
	guint32 i;
#if 0
	guint32 j;
#endif
	for(i=0; i<object->vertex_count; i++)
	{
		object->vertices[i*3+0] *= scale;
		object->vertices[i*3+1] *= scale;
		object->vertices[i*3+2] *= scale;
	}

#if 0
	for(i = 0; i < object->_num_faces; i ++)
	{
		for(j = 0; j < 6; j ++)
		{
			object->_tex_coords[i * 3 + j] *= scale;
		}
	}
#endif
}

G3DImage *free_G3DImage(G3DImage *image)
{
	if(image->pixeldata) g_free(image->pixeldata);
	if(image->name) g_free(image->name);
	g_free(image);
	return NULL;
}


G3DObject *free_G3DObject(G3DObject *object)
{
	GSList *slist;

	if(object->name != NULL) g_free(object->name);

	/* materials */
	slist = object->materials;
	while(slist != NULL)
	{
		GSList *snext;
		if(slist->data != NULL)
		{
#if 0
			G3DMaterial *mat = (G3DMaterial*)slist->data;
			/* should be attached top model->textures, so don't double-free */
			if(mat->texture != NULL)
				mat->texture = free_G3DImage(mat->texture);
			g_free(mat);
#endif
		}
		snext = slist->next;
		g_slist_free_1(slist);
		slist = snext;
	}
	/* faces */
	slist = object->faces;
	while(slist != NULL)
	{
		GSList *snext;
		G3DFace *face = (G3DFace*)slist->data;
		if(face->index_vertex != NULL) g_free(face->index_vertex);
		g_free(face);
		snext = slist->next;
		g_slist_free_1(slist);
		slist = snext;
	}
	/* vertices */
	if(object->vertices != NULL) g_free(object->vertices);
	if(object->tex_vertices != NULL) g_free(object->tex_vertices);
	if(object->_normals != NULL) g_free(object->_normals);
	if(object->_indices != NULL) g_free(object->_indices);
	if(object->_materials != NULL) g_free(object->_materials);
	if(object->_flags != NULL) g_free(object->_flags);

	g_free(object);
	return NULL;
}


void object_clearmodel(G3DModel *model)
{
	GSList *list;

	/* lights */
	list = model->lights;
	while(list != NULL)
	{
		GSList *next;
		if(list->data != NULL) g_free(list->data);
		next = list->next;
		g_slist_free_1(list);
		list = next;
	}
	model->lights = NULL;

	/* objects */
	list = model->objects;
	while(list != NULL)
	{
		GSList *next;
		G3DObject *object = (G3DObject*)list->data;

		free_G3DObject(object);

		next = list->next;
		g_slist_free_1(list);
		list = next;
	}
	model->objects = NULL;

	list = model->materials;
	while(list != NULL)
	{
		GSList *next;
		if(list->data != NULL) 
		{
#if 0
			G3DMaterial *mat = (G3DMaterial*)list->data;
			if(mat->texture != NULL)
			{
				mat->texture = free_G3DImage(mat->texture);
			}
			g_free(mat);
#endif
		}
		next = list->next;
		g_slist_free_1(list);
		list = next;
	}
	model->materials = NULL;

	if(model->filename != NULL) g_free(model->filename);
	model->filename = NULL;

	/* textures */
	g3d_texture_free_cache(model);
}

int object_centermodel(G3DModel *model)
{
	double min_x=10.0e99,min_y=10.0e99,min_z=10.0e99, 
				 max_x=-9.9e99,max_y=-9.9e99,max_z=-9.9e99;
	double off_x, off_y, off_z;
	GSList *olist = model->objects;
	while(olist != NULL)
	{
		G3DObject *object = (G3DObject*)olist->data;
		int i;
		for(i=0; i<object->vertex_count; i++)
		{
			if(object->vertices[i*3+0] < min_x) min_x = object->vertices[i*3+0];
			if(object->vertices[i*3+1] < min_y) min_y = object->vertices[i*3+1];
			if(object->vertices[i*3+2] < min_z) min_z = object->vertices[i*3+2];
			if(object->vertices[i*3+0] > max_x) max_x = object->vertices[i*3+0];
			if(object->vertices[i*3+1] > max_y) max_y = object->vertices[i*3+1];
			if(object->vertices[i*3+2] > max_z) max_z = object->vertices[i*3+2];
		}
		olist = olist->next;
	}
	off_x = max_x - ((max_x - min_x) / 2.0);
	off_y = max_y - ((max_y - min_y) / 2.0);
	off_z = max_z - ((max_z - min_z) / 2.0);
#if DEBUG > 1
	g_printerr("x: %f - %f (%-f)\n", min_x, max_x, off_x);
	g_printerr("y: %f - %f (%-f)\n", min_y, max_y, off_y);
	g_printerr("z: %f - %f (%-f)\n", min_z, max_z, off_z);
#endif
	olist = model->objects;
	while(olist != NULL)
	{
		G3DObject *object = (G3DObject*)olist->data;
		int i;
		for(i=0; i<object->vertex_count; i++)
		{
			object->vertices[i*3+0] -= off_x;
			object->vertices[i*3+1] -= off_y;
			object->vertices[i*3+2] -= off_z;
		}
		olist = olist->next;
	}
	return EXIT_SUCCESS;
}

int object_checkmodel(G3DModel *model)
{
	G3DFace *face;
	G3DObject *object;
	GSList *fitem, *oitem;
	guint32 i, no = 0, nf;

	oitem = model->objects;

	while(oitem)
	{
		object = (G3DObject *)oitem->data;

		nf = 0;
		fitem = object->faces;
		while(fitem)
		{
			face = (G3DFace *)fitem->data;

			if(face->material == NULL)
			{
				g_warning("object_check: face->material is NULL");
				return EXIT_FAILURE;
			}

			if(face->num_vertices < 3)
			{
				g_warning("object_check: face->num_vertices < 3");
				return EXIT_FAILURE;
			}

			for(i = 0; i < face->num_vertices; i ++)
			{
				if(face->index_vertex[i] >= object->vertex_count)
				{
					g_warning("object_check: [O: %d, F: %d] "
						"index_vertex[%d] >= vertex_count (%d >= %d)",
						no, nf,
						i, face->index_vertex[i], object->vertex_count);
					return EXIT_FAILURE;
				}
			}

			nf ++;
			fitem = fitem->next;
		} /* while(fitem) */

		no ++;
		oitem = oitem->next;
	} /* while(oitem) */

	return EXIT_SUCCESS;
}

int object_optimize(G3DObject *object)
{
	unsigned int num_faces = 0, index = 0;
	GSList *flist = object->faces;
	int i, j;
	GLfloat nx, ny, nz;

	/* get the number of triangles in object */
	while(flist != NULL)
	{
		G3DFace *face = (G3DFace*)flist->data;
	g_assert(face != NULL);
		if(face->num_vertices < 3)
		{
			flist = flist->next;
			continue;
		}
		num_faces += face->num_vertices - 2;
		flist = flist->next;
	}
	object->_num_faces = num_faces;
#if DEBUG > 1
	g_printerr("object_optimize: %d faces, %d triangles\n", 
						 g_slist_length(object->faces), num_faces);
#endif

	/* allocate space for triangle indices */
	object->_normals = g_malloc(num_faces * 9 * sizeof(GLfloat));
	object->_materials = g_malloc0(num_faces * sizeof(G3DMaterial*));
	object->_flags = g_malloc0(num_faces * sizeof(unsigned int));
	object->_indices = g_malloc(num_faces * 3 * sizeof(unsigned int));

	object->_tex_images = g_new0(guint32, num_faces);
	object->_tex_coords = g_new0(GLfloat, num_faces * 6);

	/* copy face data to new arrays */
	flist = object->faces;
	while(flist != NULL)
	{
		G3DFace *face = (G3DFace*)flist->data;

		object->_flags[index] = face->flags;

		if(face->num_vertices < 3)
		{
			flist = flist->next;
			continue;
		}
		/* calculate default normal */
		if(!(face->flags & G3D_FLAG_FAC_NORMALS))
		{
			GLfloat ax, ay, az, bx, by, bz;

			ax = object->vertices[face->index_vertex[1]*3+0] -
					 object->vertices[face->index_vertex[0]*3+0];
			ay = object->vertices[face->index_vertex[1]*3+1] -
					 object->vertices[face->index_vertex[0]*3+1];
			az = object->vertices[face->index_vertex[1]*3+2] -
					 object->vertices[face->index_vertex[0]*3+2];
			bx = object->vertices[face->index_vertex[2]*3+0] -
					 object->vertices[face->index_vertex[0]*3+0];
			by = object->vertices[face->index_vertex[2]*3+1] -
					 object->vertices[face->index_vertex[0]*3+1];
			bz = object->vertices[face->index_vertex[2]*3+2] -
					 object->vertices[face->index_vertex[0]*3+2];

			g3d_vector_normal(ax, ay, az, bx, by, bz, &nx, &ny, &nz);

			if(g3d_vector_unify(&nx, &ny, &nz) != EXIT_SUCCESS)
			{
				flist = flist->next;
				object->_num_faces -= face->num_vertices - 2;
				continue;
			}
		}

		for(i=0; i<(face->num_vertices - 2); i++)
		{
			object->_materials[index] = face->material;
			object->_flags[index] = face->flags;

			if(face->flags & G3D_FLAG_FAC_TEXMAP)
			{
				object->_tex_images[index] = face->tex_image->gl_texid;
			}

			for(j=0; j<3; j++)
			{
				if(j==0)
					object->_indices[index*3] = face->index_vertex[0];
				else
					object->_indices[index*3+j] = face->index_vertex[i+j];

				if(face->flags & G3D_FLAG_FAC_NORMALS)
				{
					object->_normals[(index*3+j)*3+0] =
						face->normals[(i+j)*3+0];
					object->_normals[(index*3+j)*3+1] =
						face->normals[(i+j)*3+1];
					object->_normals[(index*3+j)*3+2] =
						face->normals[(i+j)*3+2];
				}
				else
				{
					object->_normals[(index*3+j)*3+0] = nx;
					object->_normals[(index*3+j)*3+1] = ny;
					object->_normals[(index*3+j)*3+2] = nz;
				} /* END G3D_FLAG_FAC_NORMALS */

				if(face->flags & G3D_FLAG_FAC_TEXMAP)
				{
					if(j == 0)
					{
						/* u */
						object->_tex_coords[(index * 3 + j) * 2 + 0] =
							face->tex_coords[0];
						/* v */
						object->_tex_coords[(index * 3 + j) * 2 + 1] =
							face->tex_coords[1];
					}
					else
					{
						/* u */
						object->_tex_coords[(index * 3 + j) * 2 + 0] =
							face->tex_coords[(i + j) * 2 + 0];
						/* v */
						object->_tex_coords[(index * 3 + j) * 2 + 1] =
							face->tex_coords[(i + j) * 2 + 1];
					}
				} /* END G3D_FLAG_FAC_TEXMAP */
			}
			index++;
		}
		flist = flist->next;
	}

#if DEBUG > 1
	g_printerr("object_optimize: index %d, num_faces %d\n", 
		 index, object->_num_faces);
#endif
	return EXIT_SUCCESS;
}

static void normal_of_face(G3DObject *object, G3DFace *face,
	GLfloat *nx, GLfloat *ny, GLfloat *nz)
{
#define VERTEX_ITEM(a, b) object->vertices[face->index_vertex[a] * 3 + (b)]
	g3d_vector_normal(
		/* ax, ay, az */
		VERTEX_ITEM(1, 0) - VERTEX_ITEM(0, 0),
		VERTEX_ITEM(1, 1) - VERTEX_ITEM(0, 1),
		VERTEX_ITEM(1, 2) - VERTEX_ITEM(0, 2),
		/* bx, by, bz */
		VERTEX_ITEM(2, 0) - VERTEX_ITEM(0, 0),
		VERTEX_ITEM(2, 1) - VERTEX_ITEM(0, 1),
		VERTEX_ITEM(2, 2) - VERTEX_ITEM(0, 2),
		/* nx, ny, nz */
		nx, ny, nz);
#undef VERTEX_ITEM
}

int g3d_object_smooth(G3DObject *object, GLfloat max_angle)
{
	GLfloat n1x, n1y, n1z, n2x, n2y, n2z, sx, sy, sz, angle;
	guint32 nverts;
	GSList *ofitem, *ifitem;
	G3DFace *oface, *iface;
	guint32 i, j, index, ngot = 0;

#if DEBUG > 0
	g_print("object: start of smoothing (%.2f degrees)\n", max_angle);
#endif

	ofitem = object->faces;
	while(ofitem)
	{
		/* loop through all faces */
		oface = (G3DFace *)ofitem->data;

		oface->flags |= G3D_FLAG_FAC_NORMALS;
		if(oface->normals == NULL)
			oface->normals = g_new0(GLfloat, oface->num_vertices * 3);

		for(i = 0; i < oface->num_vertices; i ++)
		{
			/* loop through all vertices of face */
			normal_of_face(object, oface, &n1x, &n1y, &n1z);
			sx = n1x;
			sy = n1y;
			sz = n1z;
			nverts = 1;
			index = oface->index_vertex[i];

			ifitem = object->faces;
			while(ifitem)
			{
				iface = (G3DFace *)ifitem->data;

				if(iface == oface)
				{
					ifitem = ifitem->next;
					continue;
				}

				for(j = 0; j < iface->num_vertices; j ++)
				{
					/* try to find same vertex */
					if(iface->index_vertex[i] == index)
					{
						normal_of_face(object, iface, &n2x, &n2y, &n2z);
						if(g3d_vector_wrong(
							n1x, n1y, n1z, n2x, n2y, n2z))
						{
							n2x *= -1.0f;
							n2y *= -1.0f;
							n2z *= -1.0f;
						}
						/* angle in degrees */
						angle = g3d_vector_angle(
							n1x, n1y, n1z,
							n2x, n2y, n2z) * 180.0 / M_PI;
#if DEBUG > 0
						g_print("object: smooth: angle %.2f degrees\n",
							angle);
#endif

						if(angle <= max_angle)
						{
							sx += n2x;
							sy += n2y;
							sz += n2z;
							nverts ++;
						}
					}
				}

				ifitem = ifitem->next;
			}

			if(nverts > 1)
			{
				/* average normal */
				sx /= (GLfloat)nverts;
				sy /= (GLfloat)nverts;
				sz /= (GLfloat)nverts;
				g3d_vector_unify(&sx, &sy, &sz);
				ngot ++;
			}

			oface->normals[i * 3 + 0] = sx;
			oface->normals[i * 3 + 1] = sy;
			oface->normals[i * 3 + 2] = sz;
		} /* 0 .. num_vertices */

		ofitem = ofitem->next;
	}

#if DEBUG > 0
	g_print("object: end of smoothing (got %d)\n", ngot);
#endif

	return EXIT_SUCCESS;
}
