#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <glib.h>
#include <gmodule.h>

#include <dirent.h>
#include <sys/types.h>

#include <g3dviewer_plugin.h>

#include "plugins.h"
#include "object.h"
#include "infowin.h"

#define G3D_PLUGIN_SYMBOL(module, symbol, pointer) \
			g_module_symbol(module, symbol, (gpointer *)&(pointer))

static struct {
  GSList *import;
  GSList *image;
  GSList *unknown;
  GSList *all;
} plugins = { NULL, NULL, NULL, NULL };

static struct {
  GHashTable *import;
  GHashTable *image;
  GSList *imglist;
} extensions = { NULL, NULL, NULL };

/*****************************************************************************/

int plugins_loaddirectory(const char *dirname)
{
	DIR *plugindir;
	struct dirent *dent;
	G3DPlugin *plugin;
	gpointer gen_func;
	PluginExtensionsFunc plugin_extensions = NULL;
	PluginDescriptionFunc plugin_description = NULL;

#if DEBUG > 3
  printf("DEBUG: plugin directory: %s\n", dirname);
#endif

  plugindir = opendir(dirname);
  if(plugindir)
  {
    gint type = G3D_PLUGIN_UNKNOWN;
    gchar **last, **dir;
    gchar **dirnames = g_strsplit(dirname, "/", 0);
    dir = last = dirnames;
    while(*dir != NULL)
    {
      if(*dir != NULL) last = dir;
      dir++;
    }
#if DEBUG > 3
    fprintf(stderr, "DEBUG: last dir: %s\n", *last);
#endif
    if(strcmp("import", *last) == 0) type = G3D_PLUGIN_IMPORT;
    if(strcmp("image", *last) == 0)  type = G3D_PLUGIN_IMAGE;

    g_strfreev(dirnames);

    while((dent = readdir(plugindir)) != NULL)
    {
      if(g_strcasecmp(dent->d_name + (strlen(dent->d_name) - 3), ".so") == 0)
      {
        GModule *module;
        gchar path[2048+1];

        plugin = g_new(G3DPlugin, 1);
        plugin->name = g_strdup(dent->d_name);
        plugin->path = g_strdup(dirname);
        plugin->type = type;
        plugins.all = g_slist_append(plugins.all, plugin);
#if(DEBUG > 3)
        fprintf(stderr, "DEBUG: plugins_init: %s\n", dent->d_name);
#endif
        g_snprintf(path, 2048, "%s/%s", dirname, dent->d_name);
        module = g_module_open(path, 0);

        if(module == NULL)
        {
          fprintf(stderr, "ERROR: opening plugin '%s' failed: %s\n", 
                  dent->d_name, g_module_error());
        }
        else
        {
#if(DEBUG > 3)
          fprintf(stderr, "DEBUG: plugin_inits: %s loaded\n", dent->d_name);
#endif
          if(G3D_PLUGIN_SYMBOL(module, "plugin_description",
			gen_func) == TRUE)
          {
			plugin_description = gen_func;
            plugin->description = plugin_description();
#if DEBUG > 1
            g_print("%s:\n%s\n", plugin->name, plugin->description);
#endif
          }
          switch(plugin->type)
          {
            case G3D_PLUGIN_IMPORT: 
            {
              gchar **ext, **exts;

              plugins.import = g_slist_append(plugins.import, plugin);
              if(G3D_PLUGIN_SYMBOL(module, "plugin_extensions",
                gen_func) == TRUE)
              {
				plugin_extensions = gen_func;
                exts = plugin_extensions();
                ext = exts;
                while(*ext != NULL)
                {
                  g_hash_table_insert(extensions.import, *ext, plugin);
#if DEBUG > 3
                  fprintf(stderr, "DEBUG: import plugin '%s': extension '%s' "
                                  "registered\n", plugin->name, *ext);
#endif
                  ext++;
                }
                plugin->extensions = exts;
              }
            } break;
            case G3D_PLUGIN_IMAGE:
            {
              gchar **ext, **exts;

              plugins.image = g_slist_append(plugins.image, plugin);
              if(G3D_PLUGIN_SYMBOL(module, "plugin_extensions",
                gen_func) == TRUE)
              {
				plugin_extensions = gen_func;
                exts = plugin_extensions();
                ext = exts;
                while(*ext != NULL)
                {
                  g_hash_table_insert(extensions.image, *ext, plugin);
                  extensions.imglist = 
                    g_slist_append(extensions.imglist, *ext);
#if DEBUG > 3
                  fprintf(stderr, "DEBUG: import plugin '%s': extension '%s' "
                                  "registered\n", plugin->name, *ext);
#endif
                  ext++;
                }
                plugin->extensions = exts;
              }
            } break;
            default:
              break;
          }
          g_module_close(module);
        }
      } /* .so */
    } /* while readdir */
    closedir(plugindir);
    return EXIT_SUCCESS;
  }
  else return EXIT_FAILURE;
}

/*****************************************************************************/

int g3d_plugins_load_model(const gchar *filename, G3DModel *model)
{
  G3DPlugin *plugin;
  GSList *olist;
  GModule *module;
  char *ext, *lcext;
  char path[2048], wintitle[1024];
  PluginLoadModelFunc plugin_load;
  gpointer gen_func;
  double max_rad = 0;
#ifdef HAVE_GTK20
  gchar *basename, *dirname;
#endif

  ext = strrchr(filename, '.');
  if(ext == NULL)
  {
    g_warning("can't determine file type: %s", filename);
    return EXIT_FAILURE;
  }
  ext++;
  if(strlen(ext) == 0)
  {
    g_warning("zero length file extension: %s", filename);
    return EXIT_FAILURE;
  }
  
  lcext = g_strdup(ext);
  g_strdown(lcext);
  plugin = g_hash_table_lookup(extensions.import, lcext);
  if(plugin == NULL)
  {
    g_warning("no handler for filetype '.%s' found", lcext);
    g_free(lcext);
    return EXIT_FAILURE;
  }
  g_free(lcext);
 
  g_snprintf(path, 2048, "%s/%s", plugin->path, plugin->name);
  module = g_module_open(path, 0);
  if(module == NULL)
  {
    g_warning("can't load plugin '%s'", path);
    return EXIT_FAILURE;
  }

  if(G3D_PLUGIN_SYMBOL(module, "plugin_load", gen_func) != TRUE)
  {
    g_warning("can't find symbol 'plugin_load' in %s", path);
    g_module_close(module);
    return EXIT_FAILURE;
  }

  plugin_load = gen_func;

#ifdef HAVE_GTK20
  basename = g_path_get_basename(filename);
  dirname = g_path_get_dirname(filename);

  chdir(dirname);

  if(plugin_load(basename, model) != EXIT_SUCCESS)
#else
	if(plugin_load(filename, model) != EXIT_SUCCESS)
#endif
  {
    g_warning("plugin failed to load '%s'", filename);
    g_module_close(module);

#ifdef HAVE_GTK20
	g_free(basename);
	g_free(dirname);
#endif
    return EXIT_FAILURE;
  }
  g_module_close(module);
#ifdef HAVE_GTK20
  g_free(basename);
  g_free(dirname);
#endif
	if(object_checkmodel(model) != EXIT_SUCCESS)
	{
		object_clearmodel(model);
		return EXIT_FAILURE;
	}

  object_centermodel(model);

  olist = model->objects;
  while(olist != NULL)
  {
    G3DObject *object = (G3DObject*)olist->data;
    double radius = (double)object_radius(object);
    if(radius > max_rad) max_rad = radius;
    olist = olist->next;
  }
  olist = model->objects;
  while(olist != NULL)
  {
    G3DObject *object = (G3DObject*)olist->data;
    object_scale(object, (10.0 / max_rad));
    object_optimize(object);
    olist = olist->next;
  }

  g_snprintf(wintitle, 1024, "g3dviewer [%s]", filename);
  gtk_window_set_title(GTK_WINDOW(model->interface.window), wintitle);

  infowin_modeltab_fill(model);

	/* load textures */
	g3d_texture_load_all_textures(model);

	return EXIT_SUCCESS;
}

/*****************************************************************************/

int g3d_plugins_load_image(const gchar *filename, G3DImage *image)
{
  GModule *module;
  G3DPlugin *plugin;
  char *ext, *lcext, path[2048];
  PluginLoadImageFunc plugin_load;
  gpointer gen_func;

  ext = strrchr(filename, '.');
  if(ext == NULL)
  {
    g_warning("can't determine file type: %s", filename);
    return EXIT_FAILURE;
  }
  ext++;
  if(strlen(ext) == 0)
  {
    g_warning("zero length file extension: %s", filename);
    return EXIT_FAILURE;
  }
  
  lcext = g_strdup(ext);
  g_strdown(lcext);
  plugin = g_hash_table_lookup(extensions.image, lcext);
  if(plugin == NULL)
  {
    g_warning("no handler for filetype '.%s' found", lcext);
    g_free(lcext);
    return EXIT_FAILURE;
  }
  g_free(lcext);
 
  g_snprintf(path, 2048, "%s/%s", plugin->path, plugin->name);
  module = g_module_open(path, 0);
  if(module == NULL)
  {
    g_warning("can't load plugin '%s'", path);
    return EXIT_FAILURE;
  }

  if(G3D_PLUGIN_SYMBOL(module, "plugin_load", gen_func) != TRUE)
  {
    g_warning("can't find symbol 'plugin_load' in %s", path);
    g_module_close(module);
    return EXIT_FAILURE;
  }

  plugin_load = gen_func;

  if(plugin_load(filename, image) != EXIT_SUCCESS)
  {
    g_warning("plugin failed to load '%s'", filename);
    g_module_close(module);
    return EXIT_FAILURE;
  }
  g_module_close(module);

  return EXIT_SUCCESS;
}

/*****************************************************************************/
/* g3d_plugins_get_image_extensions:                                         */
/*   returns: NULL-terminated list of extensions                             */

gchar **g3d_plugins_get_image_extensions(void)
{
  gchar **list = g_malloc0((g_slist_length(extensions.imglist) + 1) * 
                           sizeof(gchar*));
  int pos = 0;
  GSList *lext = extensions.imglist;
  while(lext != NULL)
  {
#if DEBUG > 1
    g_printerr("g3d_plugins_get_image_extensions: %s\n", (gchar*)lext->data);
#endif
    list[pos] = (gchar*)lext->data;
    pos++;
    lext = lext->next;
  }
  return list;
}

/*****************************************************************************/

GSList *g3d_plugins_getlist(guint32 type)
{
  switch(type)
  {
    case G3D_PLUGIN_IMPORT: return plugins.import; break;
    case G3D_PLUGIN_IMAGE: return plugins.image; break;
    default: return plugins.all;
  }
}

/*****************************************************************************/

void plugins_foreach_ext(gpointer key, gpointer value, gpointer user_data)
{
  gchar *ext = (gchar*)key;
  G3DPlugin *plugin = (G3DPlugin*)value;

  printf(".%s: %s/%s\n", ext, plugin->path, plugin->name);
}

/*****************************************************************************/

int plugins_init(void)
{
  extensions.import = g_hash_table_new(g_str_hash, g_str_equal);
  extensions.image = g_hash_table_new(g_str_hash, g_str_equal);
  
  plugins_loaddirectory(PLUGIN_DIR);
  plugins_loaddirectory(PLUGIN_DIR "/image");
  plugins_loaddirectory(PLUGIN_DIR "/import");

#if DEBUG > 3
  g_hash_table_foreach(extensions.import, plugins_foreach_ext, NULL);
#endif

  return EXIT_SUCCESS;
}
